// <applet code="GameClient" width=725 height=440>
// <param name=HOST value="127.0.0.1">
// </applet>
package thegame.client;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.lang.*;
import java.net.*;
import java.text.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.border.*;


/**
 * The main class for the game client. Masters the connection to the server, the game and start screens. Provides some methods.
 * @author Erki Suurjaak
 * @version 0.99 (2000/08/28)
 */
public class GameClient
  extends JApplet
  implements GameClientConstants
{
  /** The TCP port of the server. */
  protected int
  port;

  /** The name of the local player. */
  protected String
  localPlayerName = "";

  /** The IP address of the server. */
  protected String
  host;

  /** Indicates whether the GameClient is running as an applet or as an application. */
  protected boolean
  isAnApplet = true;

  /** Contains the command-line arguments passed to the GameClient in case it was run as an application. */
  protected String[]
  mainArgs;

  /** The connection to the server. */
  protected GameClientConnection
  connection;

  /** The game screen. */
  protected GameScreen
  gameScreen;

  /** The start screen. */
  protected StartScreen
  startScreen;

  /** References the GameClient. Is used by getImageIcon(String). */
  protected static GameClient
  applet;


  /**
   * Returns an ImageIcon constructed from an image file.
   * @param path the full path and filename of the image file
   * @return an ImageIcon constructed from an image file
   */
  public static ImageIcon
  getImageIcon(final String path)
    {
      return new ImageIcon(applet.getClass().getResource(path));
    }


  /**
   * The start screen where the user can select a name and create or join a game.
   * @author Erki Suurjaak
   * @version 0.99 (2000/08/28)
   */
  protected class StartScreen
    extends JPanel
    implements GameClientConstants.StartScreenConstants
  {
    /** Represents the contents of playerNamesList. */
    protected DefaultListModel
    playerNamesListModel = new DefaultListModel();

    /** Is used to show the names of the players that can be joined. Gets its contents from playerNamesListModel. */
    protected JList
    playerNamesList = new JList(playerNamesListModel);

    /** Is used by the player to enter his/her name. */
    protected JTextField
    enterNameField = new JTextField(ENTER_NAME_FIELD_COLS);

    /** The label for enterNameField. */
    protected JLabel
    enterNameFieldLabel = new JLabel(ENTER_NAME_FIELD_LABEL_TEXT);

    /** The label that holds a picture and a copyright notice. */
    protected JLabel
    pictureAndCopyrightLabel = new JLabel(PICTURE_AND_COPYRIGHT_LABEL_TEXT,
                                          START_SCREEN_PICTURE, SwingConstants.RIGHT);

    /** The label that holds information about the game. */
    protected JLabel
    informationLabel = new JLabel(INFORMATION_LABEL_TEXT);

    /** References all the buttons. */
    protected JButton[]
    allButtons = new JButton[NUMBER_OF_BUTTONS];

    /** Is used to temporarily hold the entered player name that has been sent to the server for validation and registration. */
    protected String
    localPlayerNameToBeValidated = "";

    /** Is used to temporarily hold the name of the remote player the local player wanted to join */
    protected String
    remotePlayerNameToBeJoined = "";

    /** Is used by setNeededComponentsEnabled(boolean) to save and load the state of the components that need to be enabled or disabled. */
    protected boolean[]
    componentIsEnabled = new boolean[NUMBER_OF_COMPONENTS_TO_DISABLE];

    /** The event handler for the "Accept name" button. Sends the entered player name to the server for validation and sets certain elements of the interface disabled to disallow the player to enter a new name while the previous name is still being validated. */
    protected final ActionListener
    ON_ACCEPT_NAME_BUTTON_CLICK = new ActionListener()
      {
        public void
        actionPerformed(ActionEvent e)
          {
            String name = enterNameField.getText().trim();
            if (name.length() == 0)
              return;
            localPlayerNameToBeValidated = name;
            setNeededComponentsEnabled(false);
            tryAndSend(GameClientConnection.P2S_PLAYER_NAME_MESSAGE, name);
          }
      };

    /** The event handler for the "Create game" button. Asks the player for a symbol to be drawn (cross or circle) and brings up a GameScreen. */
    protected final ActionListener
    ON_CREATE_GAME_BUTTON_CLICK = new ActionListener()
      {
        public void
        actionPerformed(ActionEvent e)
          {
            int choice = JOptionPane.showOptionDialog(null, CREATE_GAME_BUTTON_CLICK_MESSAGE_TEXT, 
                                                      CREATE_GAME_BUTTON_CLICK_MESSAGE_TITLE,
                                                      JOptionPane.DEFAULT_OPTION, 
                                                      JOptionPane.PLAIN_MESSAGE, null, 
                                                      CREATE_GAME_BUTTON_CLICK_MESSAGE_OPTIONS,
                                                      CREATE_GAME_BUTTON_CLICK_MESSAGE_OPTIONS[0]);
            if (choice == 0)
              {
                bringUpGameScreen(true, false, null);
                tryAndSend(GameClientConnection.P2S_CREATE_GAME_MESSAGE, true);
              }
            else if (choice == 1)
              {
                bringUpGameScreen(false, false, null);
                tryAndSend(GameClientConnection.P2S_CREATE_GAME_MESSAGE, false);
              }
            else
              return;
          }
      };

    /** The event handler for the "Join game" button. Sends a join request to the server and sets certain elements of the interface disabled to disallow the player to create or join a game while the previous join request is still being handled. */
    protected final ActionListener
    ON_JOIN_GAME_BUTTON_CLICK = new ActionListener()
      {
        public void
        actionPerformed(ActionEvent e)
          {
            if (playerNamesList.isSelectionEmpty())
              return;
            if (!playerNamesList.isEnabled())
              return;
            setNeededComponentsEnabled(false);
            remotePlayerNameToBeJoined = (String)playerNamesList.getSelectedValue();
            tryAndSend(GameClientConnection.P2S_JOIN_GAME_MESSAGE, remotePlayerNameToBeJoined);
          }
      };

    /** The event handler for the "Quit game" button. Calls terminate() from GameClient. */
    protected final ActionListener
    ON_QUIT_GAME_BUTTON_CLICK = new ActionListener()
      {
        public void actionPerformed(ActionEvent e)
          {
            terminate();
          }
      };

    /** The event handler for a selection value change on playerNamesList. If a value was selected in the playerNamesList, then the "Join game" button is enabled if the user has entered a name. If a value was deselected in the playerNamesList, then the "Join game" button is disabled. */
    protected final ListSelectionListener
    ON_PLAYER_NAMES_LIST_CLICK = new ListSelectionListener()
      {
        public void valueChanged(ListSelectionEvent e)
          {
            if (localPlayerName.length() == 0)
              return;
            if (playerNamesList.isSelectionEmpty())
              allButtons[JOIN_GAME_BUTTON_NUMBER].setEnabled(false);
            else
              allButtons[JOIN_GAME_BUTTON_NUMBER].setEnabled(true);
          }
      };

    /** The event handler for a double-click on playerNamesList. Calls doClick() from the "Join game" button. */
    protected final MouseListener
    ON_PLAYER_NAMES_LIST_DOUBLE_CLICK = new MouseAdapter()
      {
         public void mouseClicked(MouseEvent e)
           {
             if (e.getClickCount() == 2)
               allButtons[JOIN_GAME_BUTTON_NUMBER].doClick();
           }
      };

    /** The event handler for a keypress on playerNamesList. If the pressed key was enter, then doClick() from the "Join game" button is called. */
    protected final KeyListener
    ON_PLAYER_NAMES_LIST_PRESS_ENTER = new KeyAdapter()
      {
        public void keyPressed(KeyEvent e)
          {
            if (e.getKeyCode() == KeyEvent.VK_ENTER)
              allButtons[JOIN_GAME_BUTTON_NUMBER].doClick();
          }
      };

    /** References the ActionListeners for all the buttons. */
    protected ActionListener[]
    BUTTON_ACTION_LISTENERS = {ON_ACCEPT_NAME_BUTTON_CLICK, ON_CREATE_GAME_BUTTON_CLICK, 
                               ON_JOIN_GAME_BUTTON_CLICK, ON_QUIT_GAME_BUTTON_CLICK, };


    /**
     * Constructs a StartScreen.
     * @param isLocalPlayerAlreadyNamed true if the player has already entered a name, false otherwise
     */
    public
    StartScreen(final boolean isLocalPlayerAlreadyNamed)
      {
        Box
        enterNamePanel = Box.createHorizontalBox(),
        gameButtonsPanel = Box.createVerticalBox(),
        leftSidePanel = Box.createVerticalBox(),
        listAndButtonsPanel = Box.createHorizontalBox(),
        rightSidePanel = Box.createVerticalBox();

        setLayout(new FlowLayout());
        for (int i = 0; i < allButtons.length; i++)
          {
            allButtons[i] = new JButton(BUTTON_LABELS[i]);
            allButtons[i].setToolTipText(BUTTON_TOOL_TIPS[i]);
            allButtons[i].setMnemonic(BUTTON_MNEMONICS[i]);
            allButtons[i].setFocusPainted(false);
            allButtons[i].setContentAreaFilled(false);
            allButtons[i].setMargin(BUTTON_INSETS);
            allButtons[i].addActionListener(BUTTON_ACTION_LISTENERS[i]);
            allButtons[i].setEnabled(ENABLED_AT_STARTUP[i]);
          }
        allButtons[CREATE_GAME_BUTTON_NUMBER].setEnabled(isLocalPlayerAlreadyNamed);
        allButtons[ACCEPT_NAME_BUTTON_NUMBER].setEnabled(!isLocalPlayerAlreadyNamed);
        for (int i = 0; i < GAME_BUTTONS.length; i++)
          gameButtonsPanel.add(allButtons[GAME_BUTTONS[i]]);

        enterNameFieldLabel.setDisplayedMnemonic(ENTER_NAME_FIELD_MNEMONIC);
        enterNameFieldLabel.setLabelFor(enterNameField);
        enterNameField.setText(localPlayerName);
        enterNameField.setEnabled(!isLocalPlayerAlreadyNamed);
        enterNamePanel.add(enterNameFieldLabel);
        enterNamePanel.add(enterNameField);
        enterNamePanel.add(allButtons[ACCEPT_NAME_BUTTON_NUMBER]);
        enterNameField.addActionListener(ON_ACCEPT_NAME_BUTTON_CLICK);

        playerNamesList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        playerNamesList.setPrototypeCellValue(PLAYER_NAMES_LIST_PROTOTYPE_VALUE);
        playerNamesList.setVisibleRowCount(PLAYER_NAMES_LIST_ROWS);
        JScrollPane playerNamesListScrollPane = new JScrollPane(playerNamesList);
        playerNamesListScrollPane.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
        playerNamesListScrollPane.setBorder(new CompoundBorder(new TitledBorder(new EtchedBorder(EtchedBorder.LOWERED),
                                                                                PLAYER_NAMES_LIST_LABEL_TEXT, 
                                                                                TitledBorder.LEFT, TitledBorder.TOP, 
                                                                                LABELS_FONT, Color.black),
                                            new CompoundBorder(new EmptyBorder(PLAYER_NAMES_LIST_INSETS),
                                                               playerNamesListScrollPane.getBorder())));
        playerNamesList.addListSelectionListener(ON_PLAYER_NAMES_LIST_CLICK);
        playerNamesList.addMouseListener(ON_PLAYER_NAMES_LIST_DOUBLE_CLICK);
        playerNamesList.addKeyListener(ON_PLAYER_NAMES_LIST_PRESS_ENTER);

        leftSidePanel.add(pictureAndCopyrightLabel);
        leftSidePanel.add(informationLabel);
        listAndButtonsPanel.add(playerNamesListScrollPane);
        listAndButtonsPanel.add(gameButtonsPanel);
        rightSidePanel.add(listAndButtonsPanel);
        rightSidePanel.add(enterNamePanel);

        add(leftSidePanel);
        add(rightSidePanel);

        tryAndSend(GameClientConnection.P2S_READY_FOR_PLAYER_LIST_MESSAGE);
      }


    /**
     * Sets certain predetermined components enabled or disabled. Used when performing functions like validating a player name etc.
     * @param enabled true if components must be enabled; false otherwise.
     */
    public void
    setNeededComponentsEnabled(final boolean enabled)
      {
        if (!enabled)
          {
            componentIsEnabled[0] = allButtons[ACCEPT_NAME_BUTTON_NUMBER].isEnabled();
            componentIsEnabled[1] = allButtons[JOIN_GAME_BUTTON_NUMBER].isEnabled();
            componentIsEnabled[2] = allButtons[CREATE_GAME_BUTTON_NUMBER].isEnabled();
            componentIsEnabled[3] = enterNameField.isEnabled();
            componentIsEnabled[4] = playerNamesList.isEnabled();
            allButtons[ACCEPT_NAME_BUTTON_NUMBER].setEnabled(false);
            allButtons[JOIN_GAME_BUTTON_NUMBER].setEnabled(false);
            allButtons[CREATE_GAME_BUTTON_NUMBER].setEnabled(false);
            enterNameField.setEnabled(false);
            playerNamesList.setEnabled(false);
          }
        else
          {
            allButtons[ACCEPT_NAME_BUTTON_NUMBER].setEnabled(componentIsEnabled[0]);
            allButtons[JOIN_GAME_BUTTON_NUMBER].setEnabled(componentIsEnabled[1]);
            allButtons[CREATE_GAME_BUTTON_NUMBER].setEnabled(componentIsEnabled[2]);
            enterNameField.setEnabled(componentIsEnabled[3]);
            playerNamesList.setEnabled(componentIsEnabled[4]);
          }
      }


    /**
     * Handles a message from the server containing the list of waiting players. Updates the contents of playerNamesList.
     * @param playerNames the player names that have been sent by the server
     */
    public void
    handlePlayerListMessage(final String[] playerNames)
      {
        String selectedValue = (String)playerNamesList.getSelectedValue();
        int selectedIndex = playerNamesList.getSelectedIndex();
        if (!playerNamesList.isSelectionEmpty())
          for (int i = 0; i < playerNames.length; i++)
            if (playerNames[i].equals(selectedValue))
              {
                selectedIndex = i;
                break;
              }
        playerNamesListModel.removeAllElements();
        if (playerNames != null)
          for (int i = 0; i < playerNames.length; i++)
            playerNamesListModel.addElement(playerNames[i]);
        if (selectedIndex != -1)
          {
            playerNamesList.setSelectedIndex(selectedIndex);
            playerNamesList.ensureIndexIsVisible(selectedIndex);
          }
      }


    /**
     * Handles a message from the server indicating that the entered name had been validated. Sets certain elements of the interface enabled to allow the player to create or join a game.
     */
    public void
    handlePlayerNameOKMessage()
      {
        localPlayerName = localPlayerNameToBeValidated;
        setNeededComponentsEnabled(true);
        enterNameField.setText(localPlayerName);
        enterNameField.setEnabled(false);
        if (!playerNamesList.isSelectionEmpty())
          allButtons[JOIN_GAME_BUTTON_NUMBER].setEnabled(true);
        allButtons[CREATE_GAME_BUTTON_NUMBER].setEnabled(true);
        allButtons[ACCEPT_NAME_BUTTON_NUMBER].setEnabled(false);
      }


    /**
     * Handles a message from the server indicating that the entered name could not be accepted. Displays errorMessage and sets certain elements of the interface enabled to allow the player to enter a new name.
     * @param errorMessage the error message sent by the server
     */
    public void
    handlePlayerNameNotOKMessage(final String errorMessage)
      {
        setNeededComponentsEnabled(true);
        Toolkit.getDefaultToolkit().beep();
        JOptionPane.showMessageDialog(null, errorMessage, PLAYER_NAME_NOT_OK_MESSAGE_TITLE, JOptionPane.PLAIN_MESSAGE);
      }


    /**
     * Handles a message from the server indicating that an attempt to join a player failed. Informs the player and sets certain elements of the interface enabled to allow the player to join or create a game.
     */
    public void
    handleJoinGameFailedMessage()
      {
        remotePlayerNameToBeJoined = "";
        setNeededComponentsEnabled(true);
        Toolkit.getDefaultToolkit().beep();
        JOptionPane.showMessageDialog(null, JOIN_GAME_FAILED_MESSAGE_TEXT, JOIN_GAME_FAILED_MESSAGE_TITLE, JOptionPane.PLAIN_MESSAGE);
      }


    /**
     * Handles a message from the server indicating that an attempt to join a player succeeded. Brings up a GameScreen.
     * @param isLocalPlayerDrawingCircles true if the player will draw circles; false if the player will draw crosses
     */
    public void
    handleJoinGameSucceededMessage(final boolean isLocalPlayerDrawingCircles)
      {
        bringUpGameScreen(isLocalPlayerDrawingCircles, true, remotePlayerNameToBeJoined);
      }
  };



  /**
   * The game screen.
   * @author Erki Suurjaak
   * @version 0.99 (2000/08/28)
   */
  protected class GameScreen
    extends JPanel
    implements GameClientConstants.GameScreenConstants
  {
    /**
     * Is used for counting the elapsed turn time of the active player; also for checking whether the local player has made a move.
     * @author Erki Suurjaak
     * @version 0.99 (2000/08/28)
     */
    protected class Monitor
      extends Thread
    {
      /** Is used to convert a time in milliseconds into a HH:MM:SS format. */
      protected SimpleDateFormat
      formatter;

      /** Is true if the Monitor is counting the time of the local player; is false if the Monitor is counting the time of the remote player. */
      protected boolean
      isTickingLocalPlayer = false;

      /** Is true if the Monitor is running; is false otherwise. */
      protected boolean
      isRunning = true;

      /** Is true if the Monitor has already been started; is false otherwise. */
      protected boolean
      isStarted = false;

      /** Is true if the Monitor has been set running but the main cycle of the run() method has not been entered yet; is false otherwise. */
      protected boolean
      isSetRunning = false;

      /** Is true if the player whose time the Monitor is counting has been changed and the Monitor is not aware of it yet; is false otherwise. */
      protected boolean
      isFirstChange = false;


      /**
       * Constructs a Monitor.
       */
      public
      Monitor()
        {
          formatter = new SimpleDateFormat ("HH:mm:ss");
          formatter.setTimeZone(TimeZone.getTimeZone("GMT"));
        }


      /**
       * Either starts or stops the Monitor thread.
       * @param running true if the thread will be set running, false if the thread will be stopped.
       */
      public void
      setRunning(final boolean running)
        {
          this.isRunning = running;
          if (isRunning) 
            {
              isSetRunning = true;
              if (!isStarted)
                {
                  isStarted = true;
                  start();
                }
            }
        }


      /**
       * Is used to change the player whose time is being counted.
       * @param isTickingLocalPlayer true if the Monitor will start counting the time of the local player; false if the Monitor will start counting the time of the remote player
       */
      public void
      setTickingLocalPlayer(final boolean isTickingLocalPlayer)
        {
          this.isTickingLocalPlayer = isTickingLocalPlayer;
          isFirstChange = true;
        }


      /**
       * Counts the time of the player whose turn it is. Also checks whether the local player has made a move and acts accordingly.
       */
      public void
      run()
        {
          long
          localPlayerCurrentTime = 0,
          localPlayerOverallTime = 0,
          remotePlayerCurrentTime = 0,
          remotePlayerOverallTime = 0,
          lastTime = 0,
          time = 0;
          boolean
          isFirstTime = true;
          while (true)
            {
              if (isSetRunning)
                {
                  isSetRunning = false;
                  localPlayerCurrentTime = 0;
                  localPlayerOverallTime = 0;
                  remotePlayerCurrentTime = 0;
                  remotePlayerOverallTime = 0;
                  lastTime = 0;
                  time = 0;
                  isFirstTime = true;
                }

              while (isRunning)
                {
                  try
                    {
                      sleep(200);
                    }
                  catch (InterruptedException e)
                    {}
                  long currentTime = System.currentTimeMillis();
                  if (isFirstTime)
                    {
                      isFirstTime = false;
                      lastTime = currentTime;
                    }

                  long elapsedTime = currentTime - lastTime;
                  time += elapsedTime;
                  lastTime = currentTime;
                  overallTimeLabel.setText(formatter.format(new Date(time)));
                  if (isTickingLocalPlayer)
                    {
                      if (isFirstChange)
                        {
                          isFirstChange = false;
                          localPlayerOverallTime -= localPlayerOverallTime % 1000;
                          localPlayerCurrentTime = time % 1000;
                          localPlayerOverallTime += localPlayerCurrentTime;
                        }
                      else
                        {
                          localPlayerCurrentTime += elapsedTime;
                          localPlayerOverallTime += elapsedTime;
                          localPlayerCurrentTimeLabel.setText(formatter.format(new Date(localPlayerCurrentTime)));
                          localPlayerOverallTimeLabel.setText(formatter.format(new Date(localPlayerOverallTime)));
                        }
                    }
                  else
                    {
                      if (isFirstChange)
                        {
                          isFirstChange = false;
                          remotePlayerOverallTime -= remotePlayerOverallTime % 1000;
                          remotePlayerCurrentTime = time % 1000;
                          remotePlayerOverallTime += remotePlayerCurrentTime;
                        }
                      else
                        {
                          remotePlayerCurrentTime += elapsedTime;
                          remotePlayerOverallTime += elapsedTime;
                          remotePlayerCurrentTimeLabel.setText(formatter.format(new Date(remotePlayerCurrentTime)));
                          remotePlayerOverallTimeLabel.setText(formatter.format(new Date(remotePlayerOverallTime)));
                        }
                    }

                  if (gameField.isChanged())
                    {
                      if (gameField.isFinished())
                        {
                          increaseLocalPlayerScore();
                          setInterfaceToGameOverMode(true);
                          tryAndSend(GameClientConnection.P2P_WINNING_MOVE_MADE_MESSAGE,
                                     gameField.getWinningSquares());
                          JOptionPane.showMessageDialog(null, PLAYER_WON_MESSAGE_TEXT,
                                                        PLAYER_WON_MESSAGE_TITLE, JOptionPane.PLAIN_MESSAGE);
                        }
                      else
                        {
                          setInterfaceToLocalPlayerMoveMode(false);
                          tryAndSend(GameClientConnection.P2P_MOVE_MADE_MESSAGE,
                                     gameField.getMove());
                        }
                    }
                }
              try
                {
                  sleep(500);
                }
              catch (InterruptedException e)
                {}
            }
        }
    }



    /** The event handler for the "New game" button. Asks the player for confirmation and, if confirmed, sends a request for a new game to the remote player. */
    protected final ActionListener
    ON_NEW_GAME_BUTTON_CLICK = new ActionListener()
      {
        public void actionPerformed(ActionEvent e)
          {
            int reply = JOptionPane.showConfirmDialog(null, NEW_GAME_BUTTON_CLICK_MESSAGE_TEXT, 
                                                      NEW_GAME_BUTTON_CLICK_MESSAGE_TITLE, 
                                                      JOptionPane.YES_NO_OPTION, JOptionPane.PLAIN_MESSAGE);
            if (reply == JOptionPane.YES_OPTION)
              {
                isRequestingForANewGame = true;
                setNeededComponentsEnabled(false);
                tryAndSend(GameClientConnection.P2P_NEW_GAME_REQUESTED_MESSAGE);
              }
          }
      };

    /** The event handler for the "Surrender" button. Asks the player for confirmation and, if confirmed, changes the interface accordingly and informs the remote player. */
    protected final ActionListener
    ON_SURRENDER_GAME_BUTTON_CLICK = new ActionListener()
      {
        public void actionPerformed(ActionEvent e)
          {
            int reply = JOptionPane.showConfirmDialog(null, SURRENDER_GAME_BUTTON_CLICK_MESSAGE_TEXT,
                                                      SURRENDER_GAME_BUTTON_CLICK_MESSAGE_TITLE, 
                                                      JOptionPane.YES_NO_OPTION, JOptionPane.PLAIN_MESSAGE);
            if (reply == JOptionPane.YES_OPTION)
              {
                increaseRemotePlayerScore();
                setInterfaceToGameOverMode(true);
                tryAndSend(GameClientConnection.P2P_PLAYER_SURRENDERED_MESSAGE);
              }
          }
      };

    /** The event handler for the "Wait for new player" button. Asks the player for confirmation and, if confirmed, drops the remote player, if any, and sets the interface accordingly to wait for a new player. */
    protected final ActionListener
    ON_WAIT_FOR_NEW_PLAYER_BUTTON_CLICK = new ActionListener()
      {
        public void
        actionPerformed(ActionEvent e)
          {
            int reply = JOptionPane.showConfirmDialog(null, WAIT_FOR_NEW_PLAYER_BUTTON_CLICK_MESSAGE_TEXT, 
                                                      WAIT_FOR_NEW_PLAYER_BUTTON_CLICK_MESSAGE_TITLE, 
                                                      JOptionPane.YES_NO_OPTION, JOptionPane.PLAIN_MESSAGE);
            if (reply == JOptionPane.YES_OPTION)
              {
                tryAndSend(GameClientConnection.P2S_PLAYER_WAITED_FOR_NEW_PLAYER_MESSAGE);
                monitor.setRunning(false);
                setPlayerIconsUnhighlighted();
                remotePlayerName = "";
                remotePlayerNameLabel.setText(REMOTE_PLAYER_START_LABEL);
                isConnectedToAPlayer = false;
                for (int i = 0; i < playerExistsButtons.length; i++)
                  playerExistsButtons[i].setEnabled(false);
                gameField.reset(false);
                gameField.setClosed();
                messageArea.setText(null);
                allButtons[CLEAR_MESSAGE_AREA_BUTTON_NUMBER].setEnabled(false);
                overallTimeLabel.setText(TIME_START_LABEL);
                localPlayerOverallTimeLabel.setText(TIME_START_LABEL);
                remotePlayerOverallTimeLabel.setText(TIME_START_LABEL);
                localPlayerCurrentTimeLabel.setText(TIME_START_LABEL);
                remotePlayerCurrentTimeLabel.setText(TIME_START_LABEL);
                localPlayerScoreLabel.setText(SCORE_START_LABEL);
                remotePlayerScoreLabel.setText(SCORE_START_LABEL);
              }
          }
      };

    /** The event handler for the "Return to start screen" button. Asks the player for confirmation and, if confirmed, informs the server and brings up a StartScreen.  */
    protected final ActionListener
    ON_QUIT_TO_START_SCREEN_BUTTON_CLICK = new ActionListener()
      {
        public void
        actionPerformed(ActionEvent e)
          {
            int reply = JOptionPane.showConfirmDialog(null, QUIT_TO_START_SCREEN_BUTTON_CLICK_MESSAGE_TEXT, 
                                                      QUIT_TO_START_SCREEN_BUTTON_CLICK_MESSAGE_TITLE, 
                                                      JOptionPane.YES_NO_OPTION, JOptionPane.PLAIN_MESSAGE);
            if (reply == JOptionPane.YES_OPTION)
              {
                tryAndSend(GameClientConnection.P2S_PLAYER_RETURNED_TO_START_SCREEN_MESSAGE);
                bringUpStartScreen(true);
              }
          }
      };

    /** The event handler for the "Quit game" button. Asks the user for confirmation and, if confirmed, calls terminate() from GameScreen. */
    protected final ActionListener
    ON_QUIT_GAME_BUTTON_CLICK = new ActionListener()
      {
        public void actionPerformed(ActionEvent e)
          {
            int reply = JOptionPane.showConfirmDialog(null, QUIT_GAME_BUTTON_CLICK_MESSAGE_TEXT, 
                                                      QUIT_GAME_BUTTON_CLICK_MESSAGE_TITLE, 
                                                      JOptionPane.YES_NO_OPTION, JOptionPane.PLAIN_MESSAGE);
            if (reply == JOptionPane.YES_OPTION)
              {
                tryAndSend(GameClientConnection.P2S_PLAYER_QUIT_MESSAGE);
                terminate();
              }
          }
      };

    /** The event handler for the "Send" button. Sends the entered message to the remote player and adds it to messageArea. */
    protected final ActionListener
    ON_SEND_MESSAGE_BUTTON_CLICK = new ActionListener()
      {
        public void
        actionPerformed(ActionEvent e)
          {
            if (!isConnectedToAPlayer)
              return;
            String message = enterMessageField.getText().trim();
            if (message.length() == 0)
              return;
            addLocalPlayerMessageToMessageArea(message);
            enterMessageField.setText(null);
            tryAndSend(GameClientConnection.P2P_TEXT_MESSAGE, message);
          }
      };

    /** The event handler for the "Clear" button. Clears messageArea. */
    protected final ActionListener
    ON_CLEAR_MESSAGE_AREA_BUTTON_CLICK = new ActionListener()
      {
        public void
        actionPerformed(ActionEvent e)
          {
            messageArea.setText(null);
            allButtons[CLEAR_MESSAGE_AREA_BUTTON_NUMBER].setEnabled(false);
          }
      };

    /** The event handler for the "Show my last move" checkbox. Calls setLocalPlayerLastSquareHighlighted(boolean) from GameField. */
    protected final ItemListener
    ON_SHOW_LOCAL_PLAYER_LAST_MOVE_CHECK_BOX_CLICK = new ItemListener()
      {
        public void
        itemStateChanged(ItemEvent e)
          {
            if (e.getStateChange() == ItemEvent.SELECTED)
              gameField.setLocalPlayerLastSquareHighlighted(true);
            else
              gameField.setLocalPlayerLastSquareHighlighted(false);
          }
      };

    /** The event handler for the "Show remote player's last move" checkbox. Calls setRemotePlayerLastSquareHighlighted(boolean) from GameField. */
    protected final ItemListener
    ON_SHOW_REMOTE_PLAYER_LAST_MOVE_CHECK_BOX_CLICK = new ItemListener()
      {
        public void
        itemStateChanged(ItemEvent e)
          {
            if (e.getStateChange() == ItemEvent.SELECTED)
              gameField.setRemotePlayerLastSquareHighlighted(true);
            else
              gameField.setRemotePlayerLastSquareHighlighted(false);
          }
      };

    /** True if the player is connected to a remote player; false otherwise. */
    protected boolean
    isConnectedToAPlayer = false;

    /** True if a game is running; false otherwise. */
    protected boolean
    isGameRunning = false;

    /** True if it's the local player's turn; false otherwise. */
    protected boolean
    isLocalPlayerMove = false;

    /** True if the local player has sent a request for a new game to the remote player; false otherwise. */
    protected boolean
    isRequestingForANewGame = false;

    /** Handles local player moves and updates the time counts. */
    protected Monitor
    monitor = new Monitor();

    /** References all the buttons. */
    protected JButton[]
    allButtons = new JButton[NUMBER_OF_BUTTONS];

    /** References buttons that are enabled only when a game is running. */
    protected JButton[]
    gameRunningButtons = new JButton[NUMBER_OF_GAME_RUNNING_BUTTONS];

    /** References buttons that are grouped together. */
    protected JButton[]
    gameStateButtons = new JButton[NUMBER_OF_GAME_STATE_BUTTONS];

    /** References buttons that are enabled only when a remote player is present. */
    protected JButton[]
    playerExistsButtons = new JButton[NUMBER_OF_PLAYER_EXISTS_BUTTONS];

    /** Shows the duration of the game in progress or of the previous game if a new one has not been started yet. */
    protected JLabel
    overallTimeLabel = new JLabel(TIME_START_LABEL, JLabel.CENTER);

    /** Shows the name of the local player. */
    protected JLabel
    localPlayerNameLabel = new JLabel(localPlayerName);

    /** Shows the duration of the local player's moves during the current game or the previous game if a new game has not been started yet. */
    protected JLabel
    localPlayerOverallTimeLabel = new JLabel(TIME_START_LABEL);

    /** Shows the duration of the local player's current turn or of the previous turn if it is not the local player's turn. */
    protected JLabel
    localPlayerCurrentTimeLabel = new JLabel(TIME_START_LABEL);

    /** Shows the score of the local player. */
    protected JLabel
    localPlayerScoreLabel = new JLabel(SCORE_START_LABEL);

    /** Shows the name of the remote player or "(NO PLAYER)" if no remote player has not joined yet or, if the remote player quit, the name of the remote player plus " (QUIT)". */
    protected JLabel
    remotePlayerNameLabel = new JLabel(REMOTE_PLAYER_START_LABEL);

    /** Shows the duration of the remote player's moves during the current game or the previous game if a new game has not been started yet. */
    protected JLabel
    remotePlayerOverallTimeLabel = new JLabel(TIME_START_LABEL);

    /** Shows the duration of the remote player's current turn or of the previous turn if it is not the remote player's turn. */
    protected JLabel
    remotePlayerCurrentTimeLabel = new JLabel(TIME_START_LABEL);

    /** Shows the score of the remote player. */
    protected JLabel
    remotePlayerScoreLabel = new JLabel(SCORE_START_LABEL);

    /** The label for enterMessageField. */
    protected JLabel
    enterMessageFieldLabel = new JLabel(ENTER_MESSAGE_FIELD_LABEL_LABEL);

    /** Holds the icon for the local player. The icon contains the symbol the local player makes (a cross or a circle) and is highlighted during the local player's turn. */
    protected JLabel
    localPlayerIcon = new JLabel();

    /** Holds the icon for the remote player. The icon contains the symbol the remote player makes (a cross or a circle) and is highlighted during the remote player's turn. */
    protected JLabel
    remotePlayerIcon = new JLabel();

    /** References labels that show the names of the players. */
    protected JLabel[]
    playerNameLabels = {localPlayerNameLabel, remotePlayerNameLabel};

    /** References labels that show the scores of the players. */
    protected JLabel[]
    playerScoreLabels = {localPlayerScoreLabel, remotePlayerScoreLabel};

    /** References labels that show the times of the players. */
    protected JLabel[]
    playerTimeLabels = {localPlayerOverallTimeLabel, localPlayerCurrentTimeLabel, remotePlayerOverallTimeLabel, remotePlayerCurrentTimeLabel};

    /** References labels that hold information about the local player. */
    protected JLabel[]
    localPlayerLabels = {localPlayerIcon, localPlayerNameLabel, localPlayerOverallTimeLabel, localPlayerCurrentTimeLabel, localPlayerScoreLabel};

    /** References labels that hold information about the remote player. */
    protected JLabel[]
    remotePlayerLabels = {remotePlayerIcon, remotePlayerNameLabel, remotePlayerOverallTimeLabel, remotePlayerCurrentTimeLabel, remotePlayerScoreLabel};

    /** References ActionListeners that will be set to the buttons. */
    protected ActionListener[]
    BUTTON_ACTION_LISTENERS = {ON_NEW_GAME_BUTTON_CLICK, ON_SURRENDER_GAME_BUTTON_CLICK, 
                               ON_WAIT_FOR_NEW_PLAYER_BUTTON_CLICK, ON_QUIT_TO_START_SCREEN_BUTTON_CLICK,
                               ON_QUIT_GAME_BUTTON_CLICK, ON_SEND_MESSAGE_BUTTON_CLICK,
                               ON_CLEAR_MESSAGE_AREA_BUTTON_CLICK};

    /** References ItemListeners that will be set to the checkboxes. */
    protected ItemListener[]
    CHECK_BOX_ITEM_LISTENERS = {ON_SHOW_LOCAL_PLAYER_LAST_MOVE_CHECK_BOX_CLICK,
                                ON_SHOW_REMOTE_PLAYER_LAST_MOVE_CHECK_BOX_CLICK};

    /** The gamefield. */
    protected GameField
    gameField;

    /** The score of the remote player. */
    protected int
    remotePlayerScore = 0;

    /** The score of the local player. */
    protected int
    localPlayerScore = 0;

    /** The name of the remote player. */
    protected String
    remotePlayerName = "";

    /** Holds the messages that have been sent between the players. */
    protected JTextArea
    messageArea = new JTextArea(MESSAGE_AREA_ROWS, MESSAGE_AREA_COLS);

    /** Accepts a message that can be sent to the remote player. */
    protected JTextField
    enterMessageField = new JTextField(SEND_MESSAGE_AREA_COLS);

    /** References all the checkboxes. */
    protected JCheckBox[]
    checkBoxes = new JCheckBox[NUMBER_OF_CHECK_BOXES];

    /** Is used to convert a time in milliseconds into a HH:mm:ss format. */
    protected SimpleDateFormat
    timeFormatter = new SimpleDateFormat("HH:mm:ss");

    /** The icon contains the symbol the local player makes (a cross or a circle). */
    protected ImageIcon
    localPlayerImageIcon;

    /** The icon contains the highlighted version of the symbol the local player makes (a cross or a circle). */
    protected ImageIcon
    localPlayerHighlightedImageIcon;

    /** The icon contains the symbol the remote player makes (a cross or a circle). */
    protected ImageIcon
    remotePlayerImageIcon;

    /** The icon contains the highlighted version of the symbol the remote player makes (a cross or a circle). */
    protected ImageIcon
    remotePlayerHighlightedImageIcon;


    /**
     * Sets certain predetermined components enabled or disabled. Used when performing functions like requesting for a new game etc.
     * @param enabled true if the components must be enabled; false otherwise
     */
    public void
    setNeededComponentsEnabled(final boolean enabled)
      {
        allButtons[SEND_MESSAGE_BUTTON_NUMBER].setEnabled(enabled);
        allButtons[NEW_GAME_BUTTON_NUMBER].setEnabled(enabled);
        if (isGameRunning)
          {
            if (isLocalPlayerMove)
              gameField.setEnabled(enabled);
            allButtons[SURRENDER_GAME_BUTTON_NUMBER].setEnabled(enabled);
          }
      }


    /**
     * Changes the highlight of players - sets one highlighted and the other unhighlighted.
     * @param highlightLocalPlayer true if the local player icon must highlighted; false if the remote player icon must be highlighted
     */
    public void
    changePlayerHighlight(boolean highlightLocalPlayer)
      {
        if (highlightLocalPlayer)
          {
            localPlayerIcon.setIcon(localPlayerHighlightedImageIcon);
            remotePlayerIcon.setIcon(remotePlayerImageIcon);
          }
        else
          {
            localPlayerIcon.setIcon(localPlayerImageIcon);
            remotePlayerIcon.setIcon(remotePlayerHighlightedImageIcon);
          }
      }


    /**
     * Unhighlights the player icons.
     */
    public void
    setPlayerIconsUnhighlighted()
      {
        localPlayerIcon.setIcon(localPlayerImageIcon);
        remotePlayerIcon.setIcon(remotePlayerImageIcon);
      }


    /**
     * Changes the interface to reflect whose move it is.
     * @param interfaceToLocalPlayerMoveMode true if the interface will be set to reflect that it is the local player's move; false if the interface will be set to reflect that it is the remote player' move
     */
    public void
    setInterfaceToLocalPlayerMoveMode(final boolean interfaceToLocalPlayerMoveMode)
      {
        isLocalPlayerMove = interfaceToLocalPlayerMoveMode;
        changePlayerHighlight(interfaceToLocalPlayerMoveMode);
        gameField.setEnabled(interfaceToLocalPlayerMoveMode);
        monitor.setTickingLocalPlayer(interfaceToLocalPlayerMoveMode);
      }


    /**
     * Sets the interface to game over mode or game running mode.
     * @param isGameOver true if the game is over; false otherwise
     */
    public void
    setInterfaceToGameOverMode(final boolean isGameOver)
      {
        for (int i = 0; i < gameRunningButtons.length; i++)
          {
             gameRunningButtons[i].setEnabled(!isGameOver);
          }
        isGameRunning = !isGameOver;
        if (isGameOver)
          {
            gameField.setClosed();
            localPlayerCurrentTimeLabel.setText(TIME_START_LABEL);
            remotePlayerCurrentTimeLabel.setText(TIME_START_LABEL);
            monitor.setRunning(false);
            setPlayerIconsUnhighlighted();
          }
        else
          {
            gameField.reset(isLocalPlayerMove);
            changePlayerHighlight(isLocalPlayerMove);
            overallTimeLabel.setText(TIME_START_LABEL);
            localPlayerOverallTimeLabel.setText(TIME_START_LABEL);
            remotePlayerOverallTimeLabel.setText(TIME_START_LABEL);
            localPlayerCurrentTimeLabel.setText(TIME_START_LABEL);
            remotePlayerCurrentTimeLabel.setText(TIME_START_LABEL);
            localPlayerScoreLabel.setText(SCORE_START_LABEL);
            remotePlayerScoreLabel.setText(SCORE_START_LABEL);
            monitor.setTickingLocalPlayer(isLocalPlayerMove);
            monitor.setRunning(true);
          }
      }


    /**
     * Increases the local player's score and updates the interface.
     */
    public void
    increaseLocalPlayerScore()
      {
        localPlayerScore++;
        localPlayerScoreLabel.setText(String.valueOf(localPlayerScore));
      }


    /**
     * Increases the remote player's score and updates the interface.
     */
    public void
    increaseRemotePlayerScore()
      {
        remotePlayerScore++;
        remotePlayerScoreLabel.setText(String.valueOf(remotePlayerScore));
      }


    /**
     * Adds toBeAdded with the remote player name and current time to messageArea. Uses addTextToMessageArea(String).
     * @param toBeAdded the text to be added
     */
    public void
    addRemotePlayerMessageToMessageArea(final String toBeAdded)
      {
        addTextToMessageArea(toBeAdded + NEW_LINE + remotePlayerName + ", "
                             + timeFormatter.format(new Date(System.currentTimeMillis()))
                             + NEW_LINE + NEW_LINE);
      }


    /**
     * Adds toBeAdded with the local player name and current time to messageArea. Uses addTextToMessageArea(String).
     * @param toBeAdded the text to be added
     */
    public void
    addLocalPlayerMessageToMessageArea(final String toBeAdded)
      {
        addTextToMessageArea(toBeAdded + NEW_LINE + localPlayerName + ", "
                             + timeFormatter.format(new Date(System.currentTimeMillis()))
                             + NEW_LINE + NEW_LINE);
      }


    /**
     * Adds toBeAdded to messageArea. The messageArea is emptied first if it has reached its capacity limit. The "Clear" button is enabled.
     * @param toBeAdded the text to be added
     */
    public void
    addTextToMessageArea(final String toBeAdded)
      {
        if (messageArea.getText().length() + toBeAdded.length() > MESSAGE_AREA_MAXIMUM_LENGTH)
          messageArea.setText(null);
        messageArea.append(toBeAdded);
        messageArea.setCaretPosition(messageArea.getText().length());
        allButtons[CLEAR_MESSAGE_AREA_BUTTON_NUMBER].setEnabled(true);
      }


    /**
     * Constructs a GameScreen. The isConnectedToAPlayer argument determines whether the player can make a move (if there is a remote player present, i.e. the local player joined somebody).
     * @param isLocalPlayerDrawingCircles true if the local player draws circles; false if the local player draws crosses
     * @param isConnectedToAPlayer true if there is a remote player present; false otherwise
     * @param remotePlayerName the name of the remote player joined, or null
     */
    public
    GameScreen(final boolean isLocalPlayerDrawingCircles, final boolean isConnectedToAPlayer,
               final String remotePlayerName)
      {
        this.isConnectedToAPlayer = isConnectedToAPlayer;
        if (remotePlayerName != null)
          this.remotePlayerName = remotePlayerName;

        JPanel
        buttonPanel = new JPanel(),
        messagesBorderedPanel = new JPanel(),
        labelsPanel = new JPanel(),
        checkBoxPanel = new JPanel();

        Box
        localPlayerLabelsPanel = Box.createVerticalBox(),
        remotePlayerLabelsPanel = Box.createVerticalBox(),
        messagesPanel = Box.createVerticalBox(),
        messagesButtonsPanel = Box.createVerticalBox(),
        messagesSendMessagePanel = Box.createVerticalBox(),
        messagesLowerPartPanel = Box.createHorizontalBox(),
        leftPartPanel = Box.createVerticalBox(),
        gameFieldAndCheckBoxesPanel = Box.createVerticalBox();

        buttonPanel.setLayout(new GridLayout(BUTTON_PANEL_ROWS, BUTTON_PANEL_COLS));
        checkBoxPanel.setLayout(new GridLayout(CHECK_BOX_PANEL_ROWS, CHECK_BOX_PANEL_COLS));
        labelsPanel.setLayout(new BorderLayout());

        gameField = new GameField(isLocalPlayerDrawingCircles);


        for (int i = 0; i < checkBoxes.length; i++)
          {
            checkBoxes[i] = new JCheckBox(CHECK_BOX_LABELS[i]);
            checkBoxes[i].setMnemonic(CHECK_BOX_MNEMONICS[i]);
            checkBoxes[i].setToolTipText(CHECK_BOX_TOOL_TIPS[i]);
            checkBoxes[i].setFocusPainted(false);
            checkBoxes[i].setContentAreaFilled(false);
            checkBoxes[i].setFont(CHECK_BOXES_FONT);
            checkBoxes[i].setMargin(CHECK_BOX_INSETS);
            checkBoxes[i].addItemListener(CHECK_BOX_ITEM_LISTENERS[i]);
            checkBoxPanel.add(checkBoxes[i]);
          }
        gameFieldAndCheckBoxesPanel.add(gameField);
        gameFieldAndCheckBoxesPanel.add(checkBoxPanel);

        for (int i = 0; i < localPlayerLabels.length; i++)
          {
            remotePlayerLabels[i].setHorizontalAlignment(JLabel.RIGHT);
            remotePlayerLabels[i].setAlignmentX(Component.LEFT_ALIGNMENT);
            localPlayerLabels[i].setFont(PLAYER_LABELS_FONT);
            localPlayerLabelsPanel.add(localPlayerLabels[i]);
          }

        for (int i = 0; i < remotePlayerLabels.length; i++)
          {
            remotePlayerLabels[i].setAlignmentX(Component.RIGHT_ALIGNMENT);
            remotePlayerLabels[i].setFont(PLAYER_LABELS_FONT);
            remotePlayerLabelsPanel.add(remotePlayerLabels[i]);
          }

        for (int i = 0; i < playerScoreLabels.length; i++)
          {
            playerScoreLabels[i].setFont(PLAYER_SCORE_LABELS_FONT);
          }

        labelsPanel.add(BorderLayout.WEST, localPlayerLabelsPanel);
        labelsPanel.add(BorderLayout.EAST, remotePlayerLabelsPanel);
        labelsPanel.add(BorderLayout.CENTER, overallTimeLabel);


        for (int i = 0; i < allButtons.length; i++)
          {
            allButtons[i] = new JButton(BUTTON_LABELS[i]);
            allButtons[i].setToolTipText(BUTTON_TOOL_TIPS[i]);
            allButtons[i].setMnemonic(BUTTON_MNEMONICS[i]);
            allButtons[i].setFocusPainted(false);
            allButtons[i].setContentAreaFilled(false);
            allButtons[i].setMargin(BUTTON_INSETS);
            allButtons[i].addActionListener(BUTTON_ACTION_LISTENERS[i]);
          }

        for (int i = 0; i < gameStateButtons.length; i++)
          {
            gameStateButtons[i] = allButtons[GAME_STATE_BUTTONS[i]];
            buttonPanel.add(gameStateButtons[i]);
          }

        for (int i = 0; i < gameRunningButtons.length; i++)
          {
            gameRunningButtons[i] = allButtons[GAME_RUNNING_BUTTONS[i]];
          }

        for (int i = 0; i < playerExistsButtons.length; i++)
          {
            playerExistsButtons[i] = allButtons[PLAYER_EXISTS_BUTTONS[i]];
          }

        allButtons[CLEAR_MESSAGE_AREA_BUTTON_NUMBER].setEnabled(false);

        allButtons[SEND_MESSAGE_BUTTON_NUMBER].setMargin(SEND_MESSAGE_BUTTON_INSETS);
  
        enterMessageField.addActionListener(ON_SEND_MESSAGE_BUTTON_CLICK);

        messageArea.setLineWrap(true);
        messageArea.setWrapStyleWord(true);

        messageArea.setEditable(false);
        JScrollPane messageAreaScrollPane = new JScrollPane(messageArea);
        messageAreaScrollPane.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
        messageAreaScrollPane.setBorder(new CompoundBorder(new EmptyBorder(MESSAGE_AREA_INSETS),
                                                           messageAreaScrollPane.getBorder()));

        if (isLocalPlayerDrawingCircles)
          {
            localPlayerImageIcon = PLAYER_CIRCLED_ICON;
            localPlayerHighlightedImageIcon = PLAYER_HIGHLIGHTED_CIRCLED_ICON;
            remotePlayerImageIcon = PLAYER_CROSSED_ICON;
            remotePlayerHighlightedImageIcon = PLAYER_HIGHLIGHTED_CROSSED_ICON;
          }
        else
          {
            localPlayerImageIcon = PLAYER_CROSSED_ICON;
            localPlayerHighlightedImageIcon = PLAYER_HIGHLIGHTED_CROSSED_ICON;
            remotePlayerImageIcon = PLAYER_CIRCLED_ICON;
            remotePlayerHighlightedImageIcon = PLAYER_HIGHLIGHTED_CIRCLED_ICON;
          }


        if (isConnectedToAPlayer)
          {
            isGameRunning = true;
            isLocalPlayerMove = true;
            changePlayerHighlight(true);
            remotePlayerNameLabel.setText(remotePlayerName);
            monitor.setTickingLocalPlayer(true);
            monitor.setRunning(true);
          }
        else
          {
            setPlayerIconsUnhighlighted();
            gameField.setClosed();
            for (int i = 0; i < playerExistsButtons.length; i++)
              playerExistsButtons[i].setEnabled(false);
          }


        enterMessageFieldLabel.setLabelFor(enterMessageField);
        enterMessageFieldLabel.setDisplayedMnemonic(SEND_MESSAGE_BUTTON_LABEL_MNEMONIC);
        messagesSendMessagePanel.add(enterMessageFieldLabel);
        messagesSendMessagePanel.add(enterMessageField);

        messagesButtonsPanel.add(allButtons[CLEAR_MESSAGE_AREA_BUTTON_NUMBER]);
        messagesButtonsPanel.add(allButtons[SEND_MESSAGE_BUTTON_NUMBER]);

        messagesLowerPartPanel.add(messagesSendMessagePanel);
        messagesLowerPartPanel.add(messagesButtonsPanel);

        messagesPanel.add(messageAreaScrollPane);
        messagesPanel.add(messagesLowerPartPanel);

        messagesBorderedPanel.setBorder(new CompoundBorder(new TitledBorder(new EtchedBorder(EtchedBorder.LOWERED),
                                                                    MESSAGE_AREA_LABEL, 
                                                                    TitledBorder.LEFT, TitledBorder.TOP, 
                                                                    PLAYER_LABELS_FONT, Color.black),
                                                   new EmptyBorder(MESSAGE_AREA_INSETS)));

        messagesBorderedPanel.add(messagesPanel);

        leftPartPanel.add(labelsPanel);
        leftPartPanel.add(messagesBorderedPanel);
        leftPartPanel.add(buttonPanel);

        add(leftPartPanel);
        add(gameFieldAndCheckBoxesPanel);
      }


    /**
     * Handles a message from the server indicating that the remote player made a move. Updates the interface accordingly.
     * @param square the coordinates of the move made
     */
    public void
    handleMoveMadeMessage(final int[] square)
      {
        if (!isConnectedToAPlayer)
          return;
        if (!isGameRunning)
          return;
        if (isLocalPlayerMove)
          return;
        if (square == null)
          return;
        gameField.setRemotePlayerClickedSquare(square);
        setInterfaceToLocalPlayerMoveMode(true);
        Toolkit.getDefaultToolkit().beep();
      }


    /**
     * Handles a message from the server indicating that the remote player made a winning move. Updates the interface accordingly and shows a message.
     * @param squares the coordinates of the winning squares
     */
    public void
    handleWinningMoveMadeMessage(final int[][] squares)
      {
        if (!isConnectedToAPlayer)
          return;
        if (!isGameRunning)
          return;
        if (isLocalPlayerMove)
          return;
        if (squares == null)
          return;
        gameField.setRemotePlayerClickedSquare(squares[0]);
        gameField.setWinningSquares(squares);
        increaseRemotePlayerScore();
        setInterfaceToGameOverMode(true);
        Toolkit.getDefaultToolkit().beep();
        JOptionPane.showMessageDialog(null, PLAYER_LOST_MESSAGE_TEXT, PLAYER_LOST_MESSAGE_TITLE,
                                      JOptionPane.PLAIN_MESSAGE);
      }


    /**
     * Handles a message from the server indicating that the remote player surrenderer. Updates the interface accordingly and shows a message.
     */
    public void
    handlePlayerSurrenderedMessage()
      {
        if (!isConnectedToAPlayer)
          return;
        if (!isGameRunning)
          return;
        increaseLocalPlayerScore();
        setInterfaceToGameOverMode(true);
        Toolkit.getDefaultToolkit().beep();
        JOptionPane.showMessageDialog(null, PLAYER_SURRENDERED_MESSAGE_TEXT,
                                      PLAYER_SURRENDERED_MESSAGE_TITLE,
                                      JOptionPane.PLAIN_MESSAGE);
      }


    /**
     * Handles a message from the server indicating that the remote player sent a message. Adds the text to messageArea.
     * @param message the text sent
     */
    public void
    handleTextMessage(final String message)
      {
        if (!isConnectedToAPlayer)
          return;
        if (message.length() == 0)
          return;
        Toolkit.getDefaultToolkit().beep();
        addRemotePlayerMessageToMessageArea(message);
      }


    /**
     * Handles a message from the server indicating that the remote player quit. Updates the interface accordingly and shows a message.
     */
    public void
    handlePlayerQuitMessage()
      {
        if (!isConnectedToAPlayer)
          return;
        isConnectedToAPlayer = false;
        for (int i = 0; i < playerExistsButtons.length; i++)
          playerExistsButtons[i].setEnabled(false);
        allButtons[WAIT_FOR_NEW_PLAYER_BUTTON_NUMBER].setEnabled(true);
        setInterfaceToGameOverMode(true);
        remotePlayerNameLabel.setText(remotePlayerName + " (QUIT)");
        remotePlayerName = "";
        Toolkit.getDefaultToolkit().beep();
        JOptionPane.showMessageDialog(null, PLAYER_QUIT_MESSAGE_TEXT, PLAYER_QUIT_MESSAGE_TITLE,
                                      JOptionPane.PLAIN_MESSAGE);
      }


    /**
     * Handles a message from the server indicating that a remote player joined the game. Updates the interface accordingly and shows a message.
     * @param playerName the name of the joining player
     */
    public void
    handlePlayerJoinedMessage(final String playerName)
      {
        if (isConnectedToAPlayer)
          return;
        remotePlayerName = playerName;
        isLocalPlayerMove = false;
        isConnectedToAPlayer = true;
        for (int i = 0; i < playerExistsButtons.length; i++)
          playerExistsButtons[i].setEnabled(true);
        remotePlayerNameLabel.setText(remotePlayerName);
        changePlayerHighlight(isLocalPlayerMove);
        setInterfaceToGameOverMode(false);
        Toolkit.getDefaultToolkit().beep();
        JOptionPane.showMessageDialog(null, NEW_PLAYER_MESSAGE_TEXT, NEW_PLAYER_MESSAGE_TITLE,
                                      JOptionPane.PLAIN_MESSAGE);
      }


    /**
     * Handles a message from the server indicating that the remote player accepted a new game. Updates the interface accordingly and shows a message.
     */
    public void
    handleNewGameAcceptedMessage()
      {
        if (!isConnectedToAPlayer)
          return;
        if (!isRequestingForANewGame)
          return;
        isRequestingForANewGame = false;
        isLocalPlayerMove = false;
        setNeededComponentsEnabled(true);
        setInterfaceToGameOverMode(false);
        Toolkit.getDefaultToolkit().beep();
        JOptionPane.showMessageDialog(null, PLAYER_ACCEPTED_NEW_GAME_MESSAGE_TEXT, 
                                      PLAYER_ACCEPTED_NEW_GAME_MESSAGE_TITLE,
                                      JOptionPane.PLAIN_MESSAGE);
      }


    /**
     * Handles a message from the server indicating that the remote player refused a new game. Updates the interface accordingly and shows a message.
     */
    public void
    handleNewGameRefusedMessage()
      {
        if (!isConnectedToAPlayer)
          return;
        if (!isRequestingForANewGame)
          return;
        isRequestingForANewGame = false;
        setNeededComponentsEnabled(true);
        Toolkit.getDefaultToolkit().beep();
        JOptionPane.showMessageDialog(null, PLAYER_REFUSED_NEW_GAME_MESSAGE_TEXT, 
                                      PLAYER_REFUSED_NEW_GAME_MESSAGE_TITLE,
                                      JOptionPane.PLAIN_MESSAGE);
      }


    /**
     * Handles a message from the server indicating that the remote player requested for a new game. Asks the local player for confirmation and, if confirmed, starts a new game with the local player making the first move. Informs the remote player.
     */
    public void
    handleNewGameRequestedMessage()
      {
        if (!isConnectedToAPlayer)
          return;
        Toolkit.getDefaultToolkit().beep();
        int reply =
          JOptionPane.showConfirmDialog(null, PLAYER_REQUESTED_NEW_GAME_MESSAGE_TEXT,
                                        PLAYER_REQUESTED_NEW_GAME_MESSAGE_TITLE,
                                        JOptionPane.YES_NO_OPTION, JOptionPane.PLAIN_MESSAGE);
        if (reply == JOptionPane.YES_OPTION)
          {
            isLocalPlayerMove = true;
            isGameRunning = true;
            tryAndSend(GameClientConnection.P2P_NEW_GAME_ACCEPTED_MESSAGE);
            setInterfaceToGameOverMode(false);
          }
        else
          tryAndSend(GameClientConnection.P2P_NEW_GAME_REFUSED_MESSAGE);
      }


    /**
     * Handles a message from the server indicating that the remote player returned to start screen. Updates the interface accordingly and shows a message.
     */
    public void
    handlePlayerReturnedToStartScreenMessage()
      {
        if (!isConnectedToAPlayer)
          return;
        isConnectedToAPlayer = false;
        for (int i = 0; i < playerExistsButtons.length; i++)
          playerExistsButtons[i].setEnabled(false);
        allButtons[WAIT_FOR_NEW_PLAYER_BUTTON_NUMBER].setEnabled(true);
        setInterfaceToGameOverMode(true);
        remotePlayerNameLabel.setText(remotePlayerName + " (QUIT)");
        remotePlayerName = "";
        Toolkit.getDefaultToolkit().beep();
        JOptionPane.showMessageDialog(null, PLAYER_QUIT_TO_START_SCREEN_MESSAGE_TEXT,
                                      PLAYER_QUIT_TO_START_SCREEN_MESSAGE_TITLE, JOptionPane.PLAIN_MESSAGE);
      }
    

    /**
     * Handles a message from the server indicating that the remote player dropped the local player. Updates the interface accordingly and shows a message.
     */
    public void
    handlePlayerDroppedMessage()
      {
        if (!isConnectedToAPlayer)
          return;
        isConnectedToAPlayer = false;
        for (int i = 0; i < playerExistsButtons.length; i++)
          playerExistsButtons[i].setEnabled(false);
        allButtons[WAIT_FOR_NEW_PLAYER_BUTTON_NUMBER].setEnabled(true);
        setInterfaceToGameOverMode(true);
        remotePlayerNameLabel.setText(REMOTE_PLAYER_START_LABEL);
        remotePlayerName = "";
        Toolkit.getDefaultToolkit().beep();
        JOptionPane.showMessageDialog(null, PLAYER_DROPPED_MESSAGE_TEXT, PLAYER_DROPPED_MESSAGE_TITLE,
                                      JOptionPane.PLAIN_MESSAGE);
      }
  };



  /**
   * Reads the HOST and PORT parameters from the applet HTML page source, or, if GameClient is run as an application, from the commandline arguments.
   */
  public void
  readParameters()
    {
      host = readParameter(HOST_PARAMETER_NAME);
      if (host == null)
        {
          try
            {
              host = getCodeBase().getHost();
            }
          catch (Exception e)
            {
              fatalError(HOST_PARAMETER_NOT_FOUND_TEXT);
            }
          if (host.length() == 0)
            fatalError(HOST_PARAMETER_NOT_FOUND_TEXT);
        }
      String s = readParameter(PORT_PARAMETER_NAME);
      try
        {
          port = Integer.parseInt(s);
        }
      catch (NumberFormatException e)
        {
          port = GameClientConnection.DEFAULT_SERVER_PORT;
        }
    }


  /**
   * Reads the required parameter from the applet HTML page source or, if GameClient is run as an application, from the commandline arguments.
   * @param parameterName the case-insensitive name of the parameter
   * @return the value of the specified parameter
   */
  public String
  readParameter(final String parameterName)
    {
      if (isAnApplet)
        return getParameter(parameterName);

      for (int i = 0; i < mainArgs.length; i += 2)
        if (mainArgs[i].equalsIgnoreCase(parameterName))
          try
            {
              return mainArgs[i + 1];
            }
          catch (Exception e)
            {
              return null;
            }
      return null;
    }


  
  /**
   * Reads the required parameters, connects to the server and brings up a StartScreen.
   */
  public void init()
    {
      readParameters();
      applet = this;

      try
        {
          connection = new GameClientConnection(new Socket(host, port), this);
        }
      catch (Exception e)
        {
          fatalError(e.toString());
        }

      bringUpStartScreen(false);
    }


  /**
   * Empties the applet window and brings up a GameScreen.
   * @param isLocalPlayerDrawingCircles true if the local player is drawing circles; false if the local player is drawing crosses
   * @param isConnectedToAPlayer true if there is a remote player present; false otherwise
   * @param remotePlayerName the name of the remote player to be joined; or null
   */
  public void
  bringUpGameScreen(final boolean isLocalPlayerDrawingCircles, final boolean isConnectedToAPlayer,
                    final String remotePlayerName)
    {
      gameScreen = new GameScreen(isLocalPlayerDrawingCircles, isConnectedToAPlayer, remotePlayerName);
      Container cp = getContentPane();
      cp.removeAll();
      cp.setLayout(new BorderLayout());
      cp.add(BorderLayout.CENTER, gameScreen);
      cp.validate();
    }


  /**
   * Empties the applet window and brings up a StartScreen.
   * @param isPlayerAlreadyNamed true if the local player already has a name; false otherwise
   */
  public void
  bringUpStartScreen(final boolean isPlayerAlreadyNamed)
    {
      startScreen = new StartScreen(isPlayerAlreadyNamed);
      Container cp = getContentPane();
      cp.removeAll();
      cp.setLayout(new BorderLayout());
      cp.add(BorderLayout.CENTER, startScreen);
      cp.validate();
    }


  /**
   * Shuts down the GameClient. Informs the server of quitting, empties the window and prints a quit message or, if GameClient is run as an application, exits. Called in case of a normal shutdown progress.
   */
  public void terminate()
    {
      try
        {
          connection.send(GameClientConnection.P2S_PLAYER_QUIT_MESSAGE);
          connection.setRunning(false);
          connection.close();
          connection = null;
        }
      catch (Exception e)
        {
        }

      if (isAnApplet)
        {
          Container cp = getContentPane();
          cp.removeAll();
          cp.setLayout(new BorderLayout());
          cp.add(BorderLayout.CENTER, new JLabel(NORMAL_TERMINATION_TEXT, JLabel.CENTER));
          cp.validate();
        }
      else
        {
          System.out.println(NORMAL_TERMINATION_TEXT);
          System.exit(0);
        }
    }


  /**
   * Shuts down the GameClient. The window is emptied and message is shown or, if GameClient is run as an application, prints message to the standard error output stream and exits. Called in case of a fatal error that does not allow to proceed normally.
   * @param message the errormessage to be shown
   */
  public void
  fatalError(final String message)
    {
      try
        {
          connection.close();
          connection = null;
        }
      catch (Exception e)
        {
        }

      if (isAnApplet)
        {
          Container cp = getContentPane();
          cp.removeAll();
          cp.setLayout(new BorderLayout());
          cp.add(BorderLayout.CENTER, new JLabel(message, JLabel.CENTER));
          cp.validate();
        }
      else
        {
          System.err.println(message);
          System.exit(1);
        }
    }


  /**
   * Calls send(int, String) from GameClientConnection. If an IOException is thrown, calls fatalError(String).
   * @param id the type identifier of the message
   * @param body the body of the message
   */
  public void
  tryAndSend(final int id, final String body)
    {
      try
        {
          connection.send(id, body);
        }
      catch (Exception e)
        {
          fatalError(e.toString());
        }
    }


  /**
   * Calls send(int, boolean) from GameClientConnection. If an IOException is thrown, calls fatalError(String).
   * @param id the type identifier of the message
   * @param body the body of the message
   */
  public void
  tryAndSend(final int id, final boolean body)
    {
      try
        {
          connection.send(id, body);
        }
      catch (Exception e)
        {
          fatalError(e.toString());
        }
    }


  /**
   * Calls send(int, int[][]) from GameClientConnection. If an IOException is thrown, calls fatalError(String).
   * @param id the type identifier of the message
   * @param body the body of the message
   */
  public void
  tryAndSend(final int id, final int[][] body)
    {
      try
        {
          connection.send(id, body);
        }
      catch (Exception e)
        {
          fatalError(e.toString());
        }
    }


  /**
   * Calls send(int, int[][]) from GameClientConnection. If an IOException is thrown, calls fatalError(String).
   * @param id the type identifier of the message
   * @param body the body of the message
   */
  public void
  tryAndSend(final int id, final int[] body)
    {
      try
        {
          connection.send(id, body);
        }
      catch (Exception e)
        {
          fatalError(e.toString());
        }
    }


  /**
   * Calls send(int) from GameClientConnection. If an IOException is thrown, calls fatalError(String).
   * @param id the type identifier of the message
   */
  public void
  tryAndSend(final int id)
    {
      try
        {
          connection.send(id);
        }
      catch (Exception e)
        {
          fatalError(e.toString());
        }
    }


  /**
   * Cleans up the data members that should be cleaned (connection to the server).
   */
  public void
  destroy()
    {
      try
        {
          connection.send(GameClientConnection.P2S_PLAYER_QUIT_MESSAGE);
          connection.setRunning(false);
        }
      catch (Exception e)
        {
        }
      finally
        {
          try
            {
              connection.close();
              connection = null;
            }
          catch (Exception e)
            {
            }
        }
    }


  /**
   * Based on the type identifier of the message, calls the needed methods from either StartScreen or GameScreen.
   * @param message the message to be handled
   */
  public synchronized void
  handleMessage(final String message)
    {
      switch (GameClientConnection.getMessageID(message))
        {
          case GameClientConnection.S2P_SERVER_SHUTDOWN_MESSAGE :
            fatalError(SERVER_SHUTDOWN_TEXT);
            break;

          case GameClientConnection.P2P_MOVE_MADE_MESSAGE :
            gameScreen.handleMoveMadeMessage(GameClientConnection.getMessageContentsAs1DIntArray(message));
            break;

          case GameClientConnection.P2P_WINNING_MOVE_MADE_MESSAGE :
            gameScreen.handleWinningMoveMadeMessage(GameClientConnection.getMessageContentsAs2DIntArray(message));
            break;

          case GameClientConnection.P2P_PLAYER_SURRENDERED_MESSAGE :
            gameScreen.handlePlayerSurrenderedMessage();
            break;

          case GameClientConnection.P2P_TEXT_MESSAGE :
            gameScreen.handleTextMessage(GameClientConnection.getMessageBody(message));
            break;

          case GameClientConnection.S2P_PLAYER_QUIT_MESSAGE :
            gameScreen.handlePlayerQuitMessage();
            break;

          case GameClientConnection.S2P_PLAYER_JOINED_MESSAGE :
            gameScreen.handlePlayerJoinedMessage(GameClientConnection.getMessageBody(message));
            break;

          case GameClientConnection.P2P_NEW_GAME_ACCEPTED_MESSAGE :
            gameScreen.handleNewGameAcceptedMessage();
            break;

          case GameClientConnection.P2P_NEW_GAME_REFUSED_MESSAGE :
            gameScreen.handleNewGameRefusedMessage();
            break;

          case GameClientConnection.S2P_PLAYER_RETURNED_TO_START_SCREEN_MESSAGE :
            gameScreen.handlePlayerReturnedToStartScreenMessage();
            break;

          case GameClientConnection.S2P_PLAYER_DROPPED_MESSAGE :
            gameScreen.handlePlayerDroppedMessage();
            break;

          case GameClientConnection.P2P_NEW_GAME_REQUESTED_MESSAGE :
            gameScreen.handleNewGameRequestedMessage();
            break;

          case GameClientConnection.S2P_PLAYER_LIST_MESSAGE :
            startScreen.handlePlayerListMessage(GameClientConnection.getMessageContentsAs1DStringArray(message));
            break;

          case GameClientConnection.S2P_PLAYER_NAME_OK_MESSAGE :
            startScreen.handlePlayerNameOKMessage();
            break;

          case GameClientConnection.S2P_PLAYER_NAME_NOT_OK_MESSAGE :
            startScreen.handlePlayerNameNotOKMessage(GameClientConnection.getMessageBody(message));
            break;

          case GameClientConnection.S2P_JOIN_GAME_FAILED_MESSAGE :
            startScreen.handleJoinGameFailedMessage();
            break;

          case GameClientConnection.S2P_JOIN_GAME_SUCCEEDED_MESSAGE :
            startScreen.handleJoinGameSucceededMessage(GameClientConnection.getMessageContentsAsBoolean(message));
            break;
        }
    }


  /**
   * Launches GameClient as an application. Possible arguments include : <BR> HOST address,<BR> where address is the IP address of the host. Is compulsory.<BR> PORT port,<BR> where port is the port of the host. Optional - if not specified, the default port will be used.
   * @param args the command-line arguments
   */
  public static void
  main(String[] args) 
    {
      GameClient gameClient = new GameClient();
      gameClient.isAnApplet = false;
      gameClient.mainArgs = args;
      JFrame frame = new JFrame("The Game");
      frame.addWindowListener(new WindowAdapter()
        {
          public void windowClosing(WindowEvent e)
            {
              System.out.println("Window Closing");
              System.exit(0);
            }
        });
      frame.getContentPane().add(gameClient);
      frame.setSize(725,440);
      gameClient.init();
      gameClient.start();
      frame.setVisible(true);
    }


}
