package thegame.client;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import javax.swing.border.*;


/**
 * The gamefield. Holds a fixed amount of GameSquares and methods to control their look and behaviour.
 * @author Erki Suurjaak
 * @version 0.99 (2000/08/28)
 */
public class GameField
  extends JPanel
{
  /** The size of a GameSquare in pixels. */
  public static final int
  SQUARE_SIZE = 15;

  /** The size of gameField. */
  public static final int
  FIELD_SIZE = 25;

  /** The minimum amount of squares in row needed for winning. */
  public static final int
  SQUARES_TO_WIN = 5;

  /** Is true if the local player is drawing circles; is false if the local player is drawing crosses. */
  protected boolean
  isDrawingCircles;

  /** Is true if the local player has made a move and the isChanged() method has not been called yet; is false otherwise. */
  protected boolean
  isChanged = false;

  /** Is true if the local player has made a winning move; is false otherwise. */
  protected boolean
  isFinished = false;

  /** Is true if the local player can make a move; is false otherwise. */
  protected boolean
  isLocalPlayerMove = true;

  /** Is true if the local player's last move is to be highlighted; is false otherwise. */
  protected boolean
  isLocalPlayerLastMoveHighlighted = false;

  /** Is true if the remote player's last move is to be highlighted; is false otherwise. */
  protected boolean
  isRemotePlayerLastMoveHighlighted = false;

  /** Holds the coordinates of the last move by the local player. */
  protected int[]
  localPlayerLastMove = {-1, -1};

  /** Holds the coordinates of the last move by the remote player. */
  protected int[]
  remotePlayerLastMove = {-1, -1};

  /** Holds the winning squares if a winning move has been made. */
  protected int[][]
  winningSquares;

  /** The gamefield. */
  protected GameSquare[][]
  gameField;


  /**
   * Sets whether or not the last square that the remote player clicked on is to be highlighted.
   * @param isHighlighted true if the last square is to be highlighted; false if the last square is not to be highlighted
   */
  public void
  setRemotePlayerLastSquareHighlighted(final boolean isHighlighted)
    {
      isRemotePlayerLastMoveHighlighted = isHighlighted;
      if (remotePlayerLastMove[0] == -1)
        return;
      gameField[remotePlayerLastMove[0]][remotePlayerLastMove[1]].setHighlighted(isHighlighted);
    }


  /**
   * Sets whether or not the last square that the local player clicked on is to be highlighted.
   * @param isHighlighted true if the last square is to be highlighted; false if the last square is not to be highlighted
   */
  public void
  setLocalPlayerLastSquareHighlighted(final boolean isHighlighted)
    {
      isLocalPlayerLastMoveHighlighted = isHighlighted;
      if (localPlayerLastMove[0] == -1)
        return;
      gameField[localPlayerLastMove[0]][localPlayerLastMove[1]].setHighlighted(isHighlighted);
    }


  /**
   * Returns the isChanged property. Sets the property false after calling.
   * @return the isChanged property
   */
  public boolean
  isChanged()
    {
      if (isChanged)
        {
          isChanged = false;
          return true;
        }
      else
        return false;
    }


  /**
   * Returns the isFinished property.
   * @return the isFinished property
   */
  public boolean
  isFinished()
    {
      return isFinished;
    }


  /**
   * Returns the coordinates of the last move made by the local player.
   * @return the coordinates of the last move
   */
  public int[]
  getMove()
    {
      return localPlayerLastMove;
    }


  /**
   * Returns the coordinates of the winning squares, or null if the local player has made no winning move.
   * @return the coordinates of the winning squares
   */
  public int[][]
  getWinningSquares()
    {
      return winningSquares;
    }

  /** The ActionListener for all the GameSquares. */
  protected final ActionListener
  ON_GAME_SQUARE_CLICK = new ActionListener()
    {
      public void actionPerformed(ActionEvent e)
        {
          if (!isLocalPlayerMove)
            return;
          isLocalPlayerMove = false;
          GameSquare clickedSquare = (GameSquare)e.getSource();
          if (!clickedSquare.isEmpty())
            return;
          if (isDrawingCircles)
            clickedSquare.setCircled();
          else
            clickedSquare.setCrossed();
          int[] clickedSquarePosition = clickedSquare.getPosition();
          if (isLocalPlayerLastMoveHighlighted)
            {
              if (localPlayerLastMove[0] != -1)
                {
                  gameField[localPlayerLastMove[0]][localPlayerLastMove[1]].setHighlighted(false);
                }
              clickedSquare.setHighlighted(true);
            }
          System.arraycopy(clickedSquarePosition, 0, localPlayerLastMove, 0, 2);
          winningSquares = findWinningSquares(clickedSquarePosition);
          if (winningSquares != null)
            {
              isFinished = true;
              setWinningSquares(winningSquares);
              setClosed();
            }
          else
            {
              setEnabled(false);
            }
          isChanged = true;
        }
    };


  /**
   * Enables or disables user input. Does not disable the squares, but modifies their behaviour to mimic disabled squares, as disabling hundreds of squares at once takes a noticeable time.
   * @param isEnabled true if user input is allowed; false if user input is forbidden
   */
  public void
  setEnabled(final boolean isEnabled)
    {
      isLocalPlayerMove = isEnabled;
      for (int i = 0; i < FIELD_SIZE; i++)
        for (int j = 0; j < FIELD_SIZE; j++)
          if (gameField[i][j].isEmpty())
            gameField[i][j].setEnabled(isEnabled);
    }


  /**
   * Closes the gamefield. Changes the icon of the empty, unclicked squares and disables user input.
   */
  public void
  setClosed()
    {
      isLocalPlayerMove = false;
      for (int i = 0; i < FIELD_SIZE; i++)
        for (int j = 0; j < FIELD_SIZE; j++)
          if (gameField[i][j].isEmpty())
            gameField[i][j].setClosed();
    }


  /**
   * Marks the specified squares as a winning row and highlights them in a special way.
   * @param squares the squares to be marked
   */
  public void
  setWinningSquares(final int[][] squares)
    {
      for (int i = 0; i < squares.length; i++)
        gameField[squares[i][0]][squares[i][1]].setWinningSquare();
      setRemotePlayerLastSquareHighlighted(isRemotePlayerLastMoveHighlighted);
      setLocalPlayerLastSquareHighlighted(isLocalPlayerLastMoveHighlighted);
    }


  /**
   * Resets the gamefield. Empties all the squares.
   * @param enabled true if user input is allowed; false if user input is forbidden
   */
  public void
  reset(final boolean enabled)
    {
      isLocalPlayerMove = enabled;
      for (int i = 0; i < FIELD_SIZE; i++)
        for (int j = 0; j < FIELD_SIZE; j++)
          gameField[i][j].reset(enabled);
      isChanged = false;
      isFinished = false;
      for (int i = 0; i < 2; i++)
        {
          localPlayerLastMove[i] = -1;
          remotePlayerLastMove[i] = -1;
        }
      winningSquares = null;
    }


  /**
   * Sets the specified square clicked by the remote player.
   * @param square the square clicked
   */
  public void
  setRemotePlayerClickedSquare(final int[] square)
    {
      if (!gameField[square[0]][square[1]].isEmpty())
        return;
      if (isRemotePlayerLastMoveHighlighted)
        if (remotePlayerLastMove[0] != -1)
          gameField[remotePlayerLastMove[0]][remotePlayerLastMove[1]].setHighlighted(false);
      remotePlayerLastMove[0] = square[0];
      remotePlayerLastMove[1] = square[1];
      if (isDrawingCircles)
          gameField[square[0]][square[1]].setCrossed();
      else
          gameField[square[0]][square[1]].setCircled();
      if (isRemotePlayerLastMoveHighlighted)
        gameField[square[0]][square[1]].setHighlighted(true);
      isLocalPlayerMove = true;
    }


  /**
   * Finds a winning row around the specified position.
   * @param position the position to be searched around
   * @return the winning row with positiom as the first element; or null if a winning row could not be found
   */
  public int[][]
  findWinningSquares(final int[] position)
    {
      final int[][] coordinateChanges =
        {
          {1, 1},
          {1, -1},
          {0, 1},
          {1, 0},
        };
      
      int[][] 
      foundSquares = new int[SQUARES_TO_WIN][2];

      int
      dx,
      dy,
      squaresInRow,
      currentSquareX,
      currentSquareY;

      boolean
      goingBackwards,
      goneBackwards;

      System.arraycopy(position, 0, foundSquares[0], 0, 2);

      for (int i = 0; i < coordinateChanges.length; i++)
        {
          dx = coordinateChanges[i][0];
          dy = coordinateChanges[i][1];
          currentSquareX = position[0];
          currentSquareY = position[1];
          squaresInRow = 1;
          goneBackwards = false;
          goingBackwards = false;

          while (true)
            {
              if (goingBackwards && !goneBackwards)
                {
                  dx = -dx;
                  dy = -dy;
                  currentSquareX = position[0];
                  currentSquareY = position[1];
                  goneBackwards = true;
                }
              currentSquareX += dx;
              currentSquareY += dy;
              if (currentSquareX < 0 || currentSquareX >= FIELD_SIZE || currentSquareY < 0 || currentSquareY >= FIELD_SIZE)
                {
                  if (goingBackwards)
                    break;
                  else
                    goingBackwards = true;
                }
              else if ((gameField[currentSquareX][currentSquareY].isCircled() && isDrawingCircles)
                  || (gameField[currentSquareX][currentSquareY].isCrossed() && !isDrawingCircles))
                {
                  foundSquares[squaresInRow][0] = currentSquareX;
                  foundSquares[squaresInRow][1] = currentSquareY;
                  squaresInRow++;
                  if (squaresInRow == SQUARES_TO_WIN)
                    return foundSquares;
                }
              else
                {
                  if (goingBackwards)
                    break;
                  else
                    goingBackwards = true;
                }
            }
        }
      return null;
    }


  /**
   * Constructs a GameField. 
   * @param isDrawingCircles true if the local player is drawing circles; false if the local player is drawing crosses
   */
  public
  GameField(final boolean isDrawingCircles)
    {
      this.isDrawingCircles = isDrawingCircles;
      gameField = new GameSquare[FIELD_SIZE][FIELD_SIZE];
      for (int i = 0; i < FIELD_SIZE; i++)
        {
          for (int j = 0; j < FIELD_SIZE; j++)
            {
              gameField[i][j] = new GameSquare(new int[] {i, j});
              gameField[i][j].addActionListener(ON_GAME_SQUARE_CLICK);
            }
        }
      Box[] bh = new Box[FIELD_SIZE];
      for (int i = 0; i < FIELD_SIZE; i++)
        {
          bh[i] = Box.createHorizontalBox();
          for (int j = 0; j < FIELD_SIZE; j++)
            {
              bh[i].add(gameField[i][j]);
            }
        }
      Box bv = Box.createVerticalBox();
      for (int i = 0; i < FIELD_SIZE; i++)
        bv.add(bh[i]);
      setLayout(new GridLayout());
      setBorder(new CompoundBorder(new EtchedBorder(EtchedBorder.LOWERED), new LineBorder(this.getBackground(), 2)));
      add(bv);
      setSize(FIELD_SIZE*SQUARE_SIZE + 6, FIELD_SIZE*SQUARE_SIZE + 6);
      return;
    }


}
