package thegame.server;

import java.net.*;
import java.io.*;
import java.util.*;
import thegame.shared.*;


/**
 * The central server that the GameClients connect to.
 * @author Erki Suurjaak
 * @version 0.99 (2000/08/28)
 */
public class GameServer
  implements GameConnectionConstants
{
  public static final int
  MAXIMUM_PLAYER_NAME_LENGTH = 25,
  MINIMUM_PLAYER_NAME_LENGTH = 2;

  public static final String[]
  FORBIDDEN_PLAYER_NAME_SUFFIXES = {"(NO PLAYER)", "(QUIT)"};

  /** The error message sent if the player entered a name of forbidden length. */
  public static final String
  PLAYER_NAME_OF_FORBIDDEN_LENGTH_MESSAGE = "The name can be " + String.valueOf(MINIMUM_PLAYER_NAME_LENGTH)
                                            + ".." + String.valueOf(MAXIMUM_PLAYER_NAME_LENGTH) + " characters long.";

  /** The error message sent if the player entered a name that contains forbidden text. */
  public static final String
  PLAYER_NAME_CONTAINS_FORBIDDEN_TEXT_MESSAGE = "The name contains forbidden text.";

  /** The error message sent if the player entered a name that is already in use. */
  public static final String
  PLAYER_NAME_IN_USE_MESSAGE = "The name is in use.";

  /** The password prompt shown at startup. */
  public static final String
  STARTUP_PASSWORD_PROMPT_MESSAGE = "Enter administrator password to be used : ";

  /** The message displayed if the administrator password was correct. */
  public static final String
  PASSWORD_CORRECT_MESSAGE = "Password correct - shutting down game server.";

  /** The message displayed if the administrator password was not correct. */
  public static final String
  PASSWORD_NOT_CORRECT_MESSAGE = "Password not correct.";

  /** The message displayed when GameServer starts. */
  public static final String
  STARTUP_MESSAGE = "Game server started on port ";

  /** Is used to hold all the ongoing GameServerConnections. */
  protected ArrayList
  connections = new ArrayList();

  /** The socket that accepts incoming connections. */
  protected ServerSocket
  serverSocket;

  /** Holds the names of the waiting players in a ready-to-send form. Is used as a cache to avoid the overhead of constructing it even when there has been no change. */
  protected String
  waitingPlayersNamesList = String.valueOf(S2P_PLAYER_LIST_MESSAGE);

  /** The administrator password. Is entered at startup and used when shutting down GameServer. */
  protected String
  administratorPassword = "";


  /**
   * Sends the names of the waiting players to all players who are able to see the list of the waiting players.
   */
  public void
  sendWaitingPlayersNames()
    {
      Iterator i = connections.iterator();
      while(i.hasNext())
        {
          GameServerConnection current = (GameServerConnection)i.next();
          if (current.isReadyForPlayerList())
            try
              {
                current.send(waitingPlayersNamesList);
              }
            catch (IOException e)
              {
              }
        }
    }



  /**
   * Updates the names of the waiting players.
   */
  public void
  updateWaitingPlayersNames()
    {
      ArrayList namesList = new ArrayList();
      Iterator i = connections.iterator();
      while(i.hasNext())
        {
          GameServerConnection current = (GameServerConnection)i.next();
          if (current.isWaiting())
            namesList.add(current.getPlayerName());
        }
      if (namesList.size() != 0)
        {
          Collections.sort(namesList);
          waitingPlayersNamesList = GameServerConnection.getReadyToSendForm(S2P_PLAYER_LIST_MESSAGE, (String[])namesList.toArray(new String[1]));
        }
      else
        waitingPlayersNamesList = String.valueOf(S2P_PLAYER_LIST_MESSAGE);
    }


  /**
   * Handles the messages sent from player to system (i.e. from GameClientConnection to GameServer).
   * @param sender the GameServerConnection that sent the message
   * @param message the message sent by the player
   */
  public void handlePlayerMessage(GameServerConnection sender, final String message)
    {
      switch (GameServerConnection.getMessageID(message))
        {
          case P2S_PLAYER_QUIT_MESSAGE :
            handlePlayerQuitMessage(sender);
            break;
          case P2S_CREATE_GAME_MESSAGE :
            handleCreateGameMessage(sender, GameServerConnection.getMessageContentsAsBoolean(message));
            break;
          case P2S_JOIN_GAME_MESSAGE :
            handleJoinGameMessage(sender, GameServerConnection.getMessageBody(message));
            break;
          case P2S_PLAYER_NAME_MESSAGE :
            handlePlayerNameMessage(sender, GameServerConnection.getMessageBody(message));
            break;
          case P2S_PLAYER_RETURNED_TO_START_SCREEN_MESSAGE :
            handlePlayerReturnedToStartScreenMessage(sender);
            break;
          case P2S_PLAYER_WAITED_FOR_NEW_PLAYER_MESSAGE :
            handlePlayerWaitedForNewPlayerMessage(sender);
            break;
          case P2S_READY_FOR_PLAYER_LIST_MESSAGE :
            handleReadyForPlayerListMessage(sender);
            break;
          default :
            return;
        }
    }


  /**
   * Handles a message to join a game.
   * @param sender the GameServerConnection that sent the message
   */
  public void
  handleJoinGameMessage(GameServerConnection sender, final String remoteName)
    {
      Iterator i = connections.iterator();
      while(i.hasNext())
        {
          GameServerConnection currentNode = (GameServerConnection)i.next();
          if (remoteName.compareToIgnoreCase(currentNode.getPlayerName()) == 0)
            {
              currentNode.setRemotePlayer(sender);
              sender.setRemotePlayer(currentNode);
              sender.setReadyForPlayerList(false);
              currentNode.setWaiting(false);
              updateWaitingPlayersNames();
              sendWaitingPlayersNames();
              try
                {
                  currentNode.send(S2P_PLAYER_JOINED_MESSAGE, sender.getPlayerName());
                  sender.send(S2P_JOIN_GAME_SUCCEEDED_MESSAGE, !currentNode.getPlayerDrawingCircles());
                }
              catch (IOException e)
                {
                }
              return;
            };
        }
      try
        {
          sender.send(S2P_JOIN_GAME_FAILED_MESSAGE);
        }
      catch (IOException e)
        {
        }
    }


  /**
   * Handles a message to create a new game.
   * @param sender the GameServerConnection that sent the message
   */
  public void
  handleCreateGameMessage(GameServerConnection sender, final boolean isPlayerDrawingCircles)
    {
      sender.setWaiting(true);
      sender.setReadyForPlayerList(false);
      sender.setPlayerDrawingCircles(isPlayerDrawingCircles);
      updateWaitingPlayersNames();
      sendWaitingPlayersNames();
    }


  /**
   * Handles a message that reports the departure of a player.
   * @param sender the GameServerConnection that sent the message
   */
  public void
  handlePlayerQuitMessage(GameServerConnection sender)
    {
      if (sender.getRemotePlayer() != null)
        {
          try
            {
              sender.getRemotePlayer().send(S2P_PLAYER_QUIT_MESSAGE);
            }
          catch (IOException e)
            {
            }
          sender.getRemotePlayer().setRemotePlayer(null);
          sender.setRemotePlayer(null);
        }
      else if (sender.isWaiting)
        {
          sender.setWaiting(false);
          updateWaitingPlayersNames();
          sendWaitingPlayersNames();
        }
      sender.setRunning(false);
      sender.close();
      connections.remove(connections.indexOf(sender));
    }


  /**
   * Handles a message that reports that a player returned to start screen
   * @param sender the GameServerConnection that sent the message
   */
  public void
  handlePlayerReturnedToStartScreenMessage(GameServerConnection sender)
    {
      sender.setReadyForPlayerList(true);
      if (sender.getRemotePlayer() != null)
        {
          try
            {
              sender.getRemotePlayer().send(S2P_PLAYER_RETURNED_TO_START_SCREEN_MESSAGE);
            }
          catch (IOException e)
            {
            }
          sender.getRemotePlayer().setRemotePlayer(null);
          sender.setRemotePlayer(null);
        }
      else if (sender.isWaiting())
        {
          sender.setWaiting(false);
          updateWaitingPlayersNames();
          sendWaitingPlayersNames();
        }
      else
        try
          {
            sender.send(waitingPlayersNamesList);
          }
        catch (IOException e)
          {
          }
    }

  
  public void
  handlePlayerWaitedForNewPlayerMessage(GameServerConnection sender)
    {
      try
        {
          sender.getRemotePlayer().send(S2P_PLAYER_DROPPED_MESSAGE);
        }
      catch (IOException e)
        {
        }
      sender.getRemotePlayer().setRemotePlayer(null);
      sender.setRemotePlayer(null);
      sender.setWaiting(true);
      updateWaitingPlayersNames();
      sendWaitingPlayersNames();
    }

  
  /**
   * Handles a message that establishes the name of a player. Tests the validity of the name (its length etc) and checks its uniqueness among the waiting players. The player is informed whether the name was acceptable or not.
   * @param sender the GameServerConnection that sent the message
   * @param name the entered name
   */
  public void
  handlePlayerNameMessage(GameServerConnection sender, final String name)
    {
      name.trim();

      try
        {
          if (name.length() > MAXIMUM_PLAYER_NAME_LENGTH
              || name.length() < MINIMUM_PLAYER_NAME_LENGTH) 
            {
              sender.send(S2P_PLAYER_NAME_NOT_OK_MESSAGE, PLAYER_NAME_OF_FORBIDDEN_LENGTH_MESSAGE);
              return;
            }
          if (name.indexOf(DELIMITER) != -1)
            {
              sender.send(S2P_PLAYER_NAME_NOT_OK_MESSAGE, PLAYER_NAME_CONTAINS_FORBIDDEN_TEXT_MESSAGE);
              return;
            }
          for (int i = 0; i < FORBIDDEN_PLAYER_NAME_SUFFIXES.length; i++)
            {
              if (name.endsWith(FORBIDDEN_PLAYER_NAME_SUFFIXES[i]))
                {
                  sender.send(S2P_PLAYER_NAME_NOT_OK_MESSAGE, PLAYER_NAME_CONTAINS_FORBIDDEN_TEXT_MESSAGE);
                  return;
                }
            }
            Iterator i = connections.iterator();
            while(i.hasNext())
              {
                GameServerConnection currentNode = (GameServerConnection)i.next();
                if (name.compareToIgnoreCase(currentNode.getPlayerName()) == 0)
                  {
                    sender.send(S2P_PLAYER_NAME_NOT_OK_MESSAGE, PLAYER_NAME_IN_USE_MESSAGE);
                    return;
                  }
              }

            sender.setPlayerName(name);
            sender.send(S2P_PLAYER_NAME_OK_MESSAGE);
        }
      catch (IOException e)
        {
        }
    }


  /**
   * Handles a message that reports that a player is ready to receive a playerlist.
   * @param sender the GameServerConnection that sent the message
   */
  public void
  handleReadyForPlayerListMessage(GameServerConnection sender)
    {
      sender.setReadyForPlayerList(true);
      try
        {
          sender.send(waitingPlayersNamesList);
        }
      catch (IOException e)
        {
        }
    }


  /**
   * Shuts down GameServer if the password is correct.
   * @param sender the GameServerConnection that sent the message
   * @param password the administrator password
   */
  public void
  handleShutdownMessage(GameServerConnection sender, final String password)
    {
      connections.remove(connections.indexOf(sender));
      try
        {
          if (password.equals(administratorPassword))
            {
              sender.send(PASSWORD_CORRECT_MESSAGE);
              sender.setRunning(false);
              sender.close();
              serverSocket.close();
              Iterator i = connections.iterator();
              while(i.hasNext())
                {
                  GameServerConnection current = (GameServerConnection)i.next();
                  current.send(S2P_SERVER_SHUTDOWN_MESSAGE);
                  current.setRunning(false);
                  current.close();
                }
              System.exit(0);
            }
          else
            {
              sender.send(PASSWORD_NOT_CORRECT_MESSAGE);
              sender.setRunning(false);
              sender.close();
            }
        }
      catch (IOException e)
        {
        }
    }


  /**
   * Constructs a GameServer listening on the specified port. The constructor prompts for an administrator password, opens a new ServerSocket on the local machine and starts waiting for connections.
   * @param port the port to be listened on
   * @throws IOException if an I/O error occurs
   */
  public
  GameServer(final int port)
  throws IOException
    {
      serverSocket = new ServerSocket(port);
      BufferedReader stdin = new BufferedReader(new InputStreamReader(System.in));
      do
        {
          System.out.print(STARTUP_PASSWORD_PROMPT_MESSAGE);
          administratorPassword = stdin.readLine().trim();
        }
      while (administratorPassword.length() == 0);
      System.out.println(STARTUP_MESSAGE + port + ".");
      while (true)
        {
          Socket socket = serverSocket.accept();
          try
            {
              GameServerConnection newConnection = new GameServerConnection(socket, this);
              connections.add(newConnection);
            }
          catch (IOException e)
            {
              socket.close();
            }
        }
    }


  /**
   * Starts up a GameServer on the local machine. The server is started on the default port or on the port that is specified in the first argument passed to the method.
   */
  public static void
  main(String[] args)
    {
      try
        {
          int port = DEFAULT_SERVER_PORT;
          if (args.length > 0)
            {
              try
                {
                  port = Integer.valueOf(args[0]).intValue();
                }
              catch (NumberFormatException e)
                {
                }
            }
          new GameServer(port);
        }
      catch (IOException e)
        {
          e.printStackTrace();
          System.exit(1);
        }
    }


}
